""" Game fix for Order of Battle: World War II
Still missing intro video codecs
"""
#pylint: disable=C0103

from protonfixes import util

def main():
    """ Install corefonts
    """

    # https://github.com/ValveSoftware/Proton/issues/639
    util.protontricks('corefonts')



""" Game fix for SUGURI 2
"""
#pylint: disable=C0103

from protonfixes import util

def main():
    """ Installs d3dxof
    """

    # https://github.com/ValveSoftware/Proton/issues/970#issuecomment-420421289
    util.protontricks('d3dxof')



""" Game fix for Titan Quest Anniversary Edition(475150)
"""
#pylint: disable=C0103

from protonfixes import util

def main():
    """ Set OS to Windows XP to pass the black menu screen
    """

    util.protontricks('winxp')



"""Game fix for Fallout 2"""

# pylint: disable=invalid-name

import sys

from protonfixes import util
from protonfixes.logger import log

def main():
    """
    Bypassing launcher, which does not run due to broken
    .NET 4.0 installation.

    Launcher provides no functionality (only a link to Bethesda site
    without any specific Fallout 2 content).
    """
    if '-classic' in sys.argv:
        log('Bypassing launcher to run Fallout 2 (Classic Version)')
        util.replace_command('Fallout2Launcher.exe', 'fallout2.exe')
    else:
        log('Bypassing launcher to run Fallout 2 (High Resolution Version)')
        util.replace_command('Fallout2Launcher.exe', 'fallout2HR.exe')



""" Run some tests and generate warnings for proton configuration issues
"""

from .logger import log

def esync_file_limits():
    """
    Check esync file limits using /proc/sys/fs/file-max
    https://www.reddit.com/r/SteamPlay/comments/9kqisk/tip_for_those_using_proton_no_esync1/
    """

    warning = '''File descriptor limit is low
    This can cause issues with ESYNC
    For more details see:
    https://github.com/zfigura/wine/blob/esync/README.esync
    '''

    with open('/proc/sys/fs/file-max') as fsmax:
        max_files = fsmax.readline()
        if int(max_files) < 8192:
            log.warn(warning)


def run_checks():
    """ Run checks to notify of any potential issues
    """

    log.info('Running checks')
    esync_file_limits()



""" Game fix for A Hat in Time
Source: https://github.com/ValveSoftware/Proton/issues/173
"""

#pylint: disable=C0103

from protonfixes import util

def main():
    """ Installs xact
    """

    util.protontricks('xact')



""" Game fix for Tesla Effect
"""
#pylint: disable=C0103

from protonfixes import util

def main():
    """ Install corefonts
    """

    # https://github.com/ValveSoftware/Proton/issues/1317
    util.protontricks('corefonts')



""" Game fix for Little Nightmares
"""
# pylint: disable=C0103

import sys
from protonfixes import util

def main():
    """ Install xact, override libraries and add launch parameter
    """

    # If not already installed, install xact
    util.protontricks('xact')

    # To fix audio crackling, set xaudio2_6.dll and xaudio2_7.dll to native
    util.winedll_override('xaudio2_6,xaudio2_7', 'n')

    # The game crashes if running with more than one CPU thread,
    # adding "-onethread" will force the game to use only one CPU thread
    sys.argv.append('-onethread')



""" Prints debug info if the environment variable DEBUG is 1
"""

import os
import sys
import shutil
# pylint: disable=E0611
from __main__ import CURRENT_PREFIX_VERSION, basedir
from .logger import log

os.environ['DEBUG'] = '1'

def show_debug_info():
    """ Show various debug info """

    check_args = [
        'iscriptevaluator.exe' in sys.argv[2],
        'getcompatpath' in sys.argv[1],
        'getnativepath' in sys.argv[1],
    ]

    if any(check_args):
        log.debug(str(sys.argv))
        return

    line = '---------------------------------------'
    log.debug('---- begin protontricks debug info ----')
    log.debug('Proton Python Version:')
    log.debug(sys.executable)
    log.debug(sys.version)
    log.debug(line)
    log.debug('System Python Version:')
    try:
        log.debug(shutil.which(os.readlink(shutil.which('python'))))
    except: #pylint: disable=W0702
        log.debug(shutil.which('python'))
    log.debug(line)

    log.debug('Proton Version:')
    log.debug(CURRENT_PREFIX_VERSION)
    log.debug(line)

    log.debug('Proton Directory:')
    log.debug(basedir)
    log.debug(line)

    ignorevars = [
        'SteamUser',
        'OLDPWD',
        'SteamAppUser',
        'LS_COLORS',
    ]

    log.debug('Environment Variables:')
    for key, value in os.environ.items():
        if key not in ignorevars:
            log.debug(key + '=' + value)
    log.debug(line)

    log.debug('Command Line:')
    log.debug(sys.argv)
    log.debug('----- end protontricks debug info -----')

show_debug_info()



""" Utilities for getting the ms corefonts for protonfixes
"""

import os
import shutil
import subprocess
import urllib.request
from threading import Thread
from tempfile import NamedTemporaryFile as Temp
from .logger import log

#pylint: disable=W0106

COREFONTS_DIR = os.path.expanduser('~/.config/protonfixes/mscorefonts/')
FONTS_LIST = [
    'andalemo.ttf', 'arial.ttf', 'arialbd.ttf',
    'arialbi.ttf', 'ariali.ttf', 'ariblk.ttf',
    'comic.ttf', 'comicbd.ttf', 'cour.ttf',
    'courbd.ttf', 'courbi.ttf', 'couri.ttf',
    'georgia.ttf', 'georgiab.ttf', 'georgiai.ttf',
    'georgiaz.ttf', 'impact.ttf', 'times.ttf',
    'timesbd.ttf', 'timesbi.ttf', 'timesi.ttf',
    'trebuc.ttf', 'trebucbd.ttf', 'trebucbi.ttf',
    'trebucit.ttf', 'verdana.ttf', 'verdanab.ttf',
    'verdanai.ttf', 'verdanaz.ttf', 'webdings.ttf',
    ]


def check_corefonts():
    """ Returns True if all corefonts installed, else returns False
    """

    if not os.path.isdir(COREFONTS_DIR):
        return False

    installed = os.listdir(COREFONTS_DIR)
    fonts = [False for x in FONTS_LIST if x not in installed]
    if all(fonts):
        return True
    return False


def download_file(url, files):
    """ Downloads a file at url and returns the adds the filename to files
    """

    log.debug('Downloading ' + url)
    with urllib.request.urlopen(url, files) as font:
        with Temp(delete=False, prefix='font' + str(os.getpid())) as temp:
            shutil.copyfileobj(font, temp)
            files.append(temp)


def extract_cab(filename, dst_dir):
    """ Extract a cab file filename to dst_dir using cabextract or bsdtar
    """

    log.debug('Extracting ' + filename)
    cabextract = shutil.which('cabextract')
    bsdtar = shutil.which('bsdtar')

    if cabextract:
        # cabextract command to extract just ttf files
        cabextract_cmd = [cabextract,
                          '-L', # make filenames lowercase
                          '-F', '*.ttf', # only extract ttf
                          '-d', dst_dir, # extract to dst_dir
                          filename,
                          ]
        cab = subprocess.Popen(cabextract_cmd)
        cab.wait()
        return True

    if bsdtar:
        # bsdtar command to extract just ttf files
        bsdtar_cmd = [bsdtar,
                      '-C', dst_dir,
                      '-xf', filename,
                      '*.[Tt][Tt][Ff]', # just TTF files
                      ]

        cab = subprocess.Popen(bsdtar_cmd)
        cab.wait()

        return True
    return False


def get_corefonts():
    """ Downloads the ms-corefonts from the pushcx/corefonts mirror
    """

    if check_corefonts():
        return

    urlbase = 'https://github.com/pushcx/corefonts/raw/master/'
    urlsuffix = '32.exe'
    fonts = ['andale', 'arial', 'arialb', 'comic', 'courie', 'georgi',
             'impact', 'times', 'trebuc', 'verdan', 'webdin', ]

    try:
        os.makedirs(COREFONTS_DIR)
    except FileExistsError:
        log.debug('Directory exists: ' + COREFONTS_DIR)

    urls = [urlbase + x + urlsuffix for x in fonts]
    files = []

    # download files in threads
    dl_threads = [Thread(target=download_file, args=(u, files))
                  for u in urls]
    [thread.start() for thread in dl_threads]
    [thread.join() for thread in dl_threads]

    # extract files with threads
    ex_threads = [Thread(target=extract_cab, args=(f.name, COREFONTS_DIR))
                  for f in files]
    [thread.start() for thread in ex_threads]
    [thread.join() for thread in ex_threads]

    # rename files to lowercase
    [shutil.move(os.path.join(d, f), os.path.join(d, f).lower())
     for d, _, files in os.walk(COREFONTS_DIR)
     for f in files]

    # cleanup temp files
    [log.debug('Removing ' + f.name) for f in files]
    [os.remove(f.name) for f in files]


def link_fonts(directory):
    """ Make links to core fonts in directory
    """

    log.info('Creating MS Core font links in ' + directory)
    src = COREFONTS_DIR
    dst = directory
    [os.link(os.path.join(d, f), os.path.join(dst, f))
     for d, _, files in os.walk(src)
     for f in files]



""" Game fix for Rise of Nations: Extended Edition
Source: https://github.com/simons-public/protonfixes/issues/24#issue-372384148
"""

#pylint: disable=C0103

from protonfixes import util

def main():
    """ Installs crypt32 and and bypasses launcher
    """

    # Install crypt32 (not required for Proton 3.16-3 and up)
    util.protontricks('crypt32')

    # Install directmusic, set overrides
    util.protontricks('directmusic')
    util.winedll_override('dmime', 'n')
    util.winedll_override('dmsynth', 'n')
    util.winedll_override('dmusic', 'n')
    util.winedll_override('dsound', 'n')
    util.winedll_override('dswave ', 'n')
    util.winedll_override('l3codecx', 'n')

    # Set sound to alsa
    util.protontricks('sound=alsa')

    # Disable launcher
    util.replace_command('patriots.exe', 'riseofnations.exe')



""" Gets the game id and applies a fix if found
"""

from __future__ import print_function
import os
import re
import sys
from importlib import import_module
from protonfixes.splash import splash
from .corefonts import check_corefonts, get_corefonts, link_fonts
from .util import protonprefix
from .checks import run_checks
from .logger import log

def game_id():
    """ Trys to return the game id from environment variables
    """

    if 'SteamAppId' in os.environ:
        return os.environ['SteamAppId']
    if 'SteamGameId' in os.environ:
        return os.environ['SteamGameId']
    if 'STEAM_COMPAT_DATA_PATH' in os.environ:
        return re.findall(r'\d+', os.environ['STEAM_COMPAT_DATA_PATH'])[-1]

    log.crit('Game ID not found in environment variables')
    return None


def game_name():
    """ Trys to return the game name from environment variables
    """

    game_library = re.findall(r'.*/steamapps', os.environ['PWD'], re.IGNORECASE)[-1]
    game_manifest = os.path.join(game_library, 'appmanifest_' + game_id() + '.acf')
    try:
        with open(game_manifest, 'r') as appmanifest:
            for xline in appmanifest.readlines():
                if 'name' in xline.strip():
                    name = re.findall(r'"[^"]+"', xline)[-1]
                    return name
    except OSError:
        return 'UNKNOWN'
    return 'UNKNOWN'


def run_fix(gameid):
    """ Loads a gamefix module by it's gameid
    """

    if gameid is None:
        return

    game = game_name() + ' ('+ gameid + ')'
    localpath = os.path.expanduser('~/.config/protonfixes/localfixes')
    if os.path.isfile(os.path.join(localpath, gameid + '.py')):
        open(os.path.join(localpath, '__init__.py'), 'a').close()
        sys.path.append(os.path.expanduser('~/.config/protonfixes'))
        try:
            game_module = import_module('localfixes.' + gameid)
            log.info('Using local protonfix for ' + game)
            game_module.main()
        except ImportError:
            log.info('No local protonfix found for ' + game)
    else:
        try:
            game_module = import_module('protonfixes.gamefixes.' + gameid)
            log.info('Using protonfix for ' + game)
            game_module.main()
        except ImportError:
            log.info('No protonfix found for ' + game)

    # install corefonts
    fontsdir = os.path.join(protonprefix(), 'drive_c/windows/Fonts')
    try:
        os.makedirs(fontsdir)
    except FileExistsError:
        log.debug('Fonts directory exists')
    if len(os.listdir(fontsdir)) < 30:
        link_fonts(fontsdir)



def main():
    """ Runs the gamefix, with splash if zenity or cefpython3 is available
    """

    check_args = [
        'iscriptevaluator.exe' in sys.argv[2],
        'getcompatpath' in sys.argv[1],
        'getnativepath' in sys.argv[1],
    ]

    if any(check_args):
        log.debug(str(sys.argv))
        log.debug('Not running protonfixes for setup runs')
        return

    with splash():
        log.info('Running protonfixes')
        run_checks()
        if not check_corefonts():
            log.info('Getting ms-corefonts')
            get_corefonts()
        run_fix(game_id())
